import duckdb
import pyarrow as pa
import pyarrow.compute as pc

# Local imports
from .ml.data import _is_numeric
from .wrangle import _clean_name

numeric_types = {
    pa.int8(), 
    pa.int16(),
    pa.int32(),
    pa.int64(),
    pa.float16(),
    pa.float32(),
    pa.float64(),
    pa.bool_()
}

categorical_types = {
    pa.int8(), 
    pa.int16(),
    pa.int32(),
    pa.int64(),
    pa.string(),
    pa.bool_()
}

def featurize(
    event_col,
    join_field,
    time_field,
    wrangled_table,
    time_period,
    target_column,
    target_value=None,
    events=None,
    histogram_top_k=50,
    histogram_cols=None,
    top_k=10,
    drop_null=True,
    n_time_periods_per_row_for_features=1,
    n_time_periods_per_row_for_target=1,
    filter_inactive=False,
    binarize_target=False,
    binarize_threshold=0):
    """
    Featurize a wrangled event table. Compute features for each entity
    in `join_field` for each `time_period`. For numerical columns, sum,
    average, and variance are computed for the current time period and previous
    `n_time_periods_per_row_for_features` periods. For categorical columns,
    the count of distinct values and the mode value are computed for the
    current time period and previous periods.
    For the categorical columns in `histogram_cols`, features are created for the
    counts of the specified values or of the top `histogram_top_k` most frequent
    values in the columns. Finally, a target is computed for each row by summing the
    value of `target_column` for the next `n_time_periods_per_row_for_target` periods if
    `target_column` is numerical, otherwise by summing the count of the `target_value`
    value within `target_column` for the next periods.

    Parameters
    __________
    event_col : str
        Name of column containing the type of each event.

    join_field : str
        Name of column containing logical entities to group by (e.g. user ID column).

    time_field : str
        Name of column containing event timestamp.

    wrangled_table : pyarrow.Table
        A merged event table returned by drifto.wrangle().

    time_period : str
        The desired time period over which to compute features, one of 'day', 'week', or
        'month'.

    target_column : str
        Name of column to use for target computation.

    target_value : str, default None
        If `target_column` is categorical, the value in `target_column` to count to compute
        the target.

    events : list of str, default None
        List of values `event_col` column whose counts to turn into features. If None,
        simply use the `histogram_top_k` most frequent values of `event_col`.

    histogram_top_k : int, default 50
        For `event_col` column and `histogram_cols` columns with no explicit value list,
        the counts of the `histogram_top_k` most frequent values are turned into features.

    histogram_cols : list of str or tuple, default None
        Names of columns whose value counts to turn into features. List item can also be a
        tuple with (column name, whitelist of values whose counts to turn into features)
        if the automatic behavior of using the `histogram_top_k` most frequent values is not
        desired.

    top_k : int, default 10
        Number of distinct values of categorical columns (such as mode) in the feature table
        to retain. All other values are replaced with a placeholder like '_OTHER_'.

    drop_null : bool, default True
        Drop rows in the feature table with null features.

    n_time_periods_per_row_for_features : int, default 1
        Number of previous periods' features to include in the current row.

    n_time_periods_per_row_for_target : int, default 1
        Number of future periods' `target_column` values or value counts to sum to get
        the target for the current row.

    filter_inactive : bool, default False
        Only return rows in the feature and inference tables where target is nonzero
        for at least one of the current period and last `n_time_periods_per_row_for_features`.

    binarize_target : bool, default False
        Binarize the target variable by setting it to 1 if its value is greater than
        `binarize_threshold`, and 0 otherwise.

    binarize_threshold : int, default 0
        Threshold for target binarization.

    Returns
    _______
    Feature table (pyarrow.Table), inference table (pyarrow.Table), and metadata (dict) about
    each feature (used by training and inference). The inference table
    contains the rows for the most recent time period in the event table, which can be used
    to make forward-looking predictions.
    """

    if histogram_cols == None:
        histogram_cols = []

    # Roll up important cols/fields into one tuple
    fields = (event_col, join_field, time_field)

    # Do some initial booking to track col types
    num_cols = []
    cat_cols = []
    for field in wrangled_table.schema:
        if field.name == join_field:
            continue
        if field.type in categorical_types:
            cat_cols.append(field.name)
        if field.type in numeric_types:
            num_cols.append(field.name)

    # Add duckdb connection for fast in-memory SQL
    con = duckdb.connect(database=':memory:')

    # Tally up the total number of time windows and cast time to microsecs 
    wrangled_table, tot_num_time_periods, start_time = _process_tot_num_time_periods(
        wrangled_table, time_field, time_period)
    wrangled_table = wrangled_table.set_column(
        wrangled_table.schema.get_field_index(time_field), time_field,
        pc.cast(wrangled_table[time_field], pa.timestamp('us')))

    # Put in dummy events for correct time splitting behavior
    dummy_events = _create_dummy_events(con, wrangled_table,
        join_field, time_field, time_period, start_time.strftime("%Y-%m-%d"),
        tot_num_time_periods)
    wrangled_table = pa.concat_tables([wrangled_table, dummy_events],
        promote=True)

    # Process heavy cols
    histogram_cols = map(lambda c: _clean_name(c) if '->' in c else c, histogram_cols)
    heavy_cat_spec = [(event_col, events)]
    for h in histogram_cols:
        heavy_cat_spec.append(h if type(h) == tuple else (h, None))
    heavy_cat_cols = []
    for col, vals in heavy_cat_spec:
        if vals == None: # User has not provided value whitelist
            p_vals = _process_top_k(histogram_top_k if col == event_col else top_k,
                col, wrangled_table, con)
            vals = [str(val) for val in p_vals[col] if val.as_py() != None]
        heavy_cat_cols.append((col, vals))

    # Core feature computation routine 
    feature_table = _compute_features(con, wrangled_table,
        join_field, time_field, heavy_cat_cols, cat_cols, num_cols,
        n_time_periods_per_row_for_features)

    # Apply top k to categorical (str) feature cols
    for S in feature_table.schema:
        if S.type == pa.string() and S.name != join_field:
            vals = _process_top_k(top_k, S.name, feature_table, con)
            vals = [str(val) for val in vals[S.name]]
            filtered_col = _process_cat_value_filter(
                vals, S.name, feature_table, con)
            idx = feature_table.column_names.index(S.name)
            feature_table = feature_table.set_column(
                idx, S.name, filtered_col[S.name])

    # Compute the target column
    labels = _compute_labels(con, feature_table, join_field, target_column, [target_value],
        n_time_periods_per_row_for_target)

    # Merge feature and target columns
    feature_table = con.execute(f"""SELECT feature_table.*, labels.label FROM feature_table JOIN labels ON
                          feature_table.{join_field} = labels.{join_field} AND feature_table.time_period = labels.time_period
                       """).arrow()
    if filter_inactive:
        feature_table = _filter_inactive(feature_table, event_col, target_value,
            n_time_periods_per_row_for_features=n_time_periods_per_row_for_features)

    # Compute inference table
    max_date = pc.max(feature_table['time_period']).as_py().strftime("%Y-%m-%d")
    inf_table = con.execute(
        f"SELECT * from feature_table WHERE time_period = '{max_date}'").arrow()

    if binarize_target:
        new_label = pc.greater(feature_table['label'], binarize_threshold)
        feature_table = feature_table.drop(['label']).append_column('label', new_label)

    # Drop missing labels in feature_table (those are inference points)
    if drop_null:
        feature_table = con.execute(
            f"""
                SELECT * FROM feature_table 
                WHERE feature_table.label IS NOT NULL;
            """).arrow()

    return (feature_table, inf_table,
            _compute_metadata(feature_table, fields, con))

def _filter_inactive(feature_table, event_col, target_value, min_active_time_periods=1,
    n_time_periods_per_row_for_features=1):

    col_names = _gen_col_names(event_col, [target_value])
    time_periods = []
    for i in range(n_time_periods_per_row_for_features + 1):
        suffix = "" if i == 0 else f"_{i}"
        agg = ' + '.join([f"COALESCE({c}_count{suffix}, 0)" for c in col_names])
        time_periods.append(f"(CASE WHEN {agg} > 0 THEN 1 ELSE 0 END)")
    time_periods = ' + '.join(time_periods)
    con = duckdb.connect(database=':memory:')
    return con.execute(f"SELECT * FROM feature_table WHERE {time_periods} >= {min_active_time_periods}").arrow()

def _gen_col_names(col_name, val_list):
    # TODO: Rename this and clean up naming notation
    return list(map(lambda e: col_name + '_' + _clean_name(e), val_list))

def _process_tot_num_time_periods(wrangled_table, time_field, time_period):
    floored_dates = pc.floor_temporal(
        wrangled_table[time_field], unit=time_period)
    wrangled_table = wrangled_table.set_column(
        wrangled_table.schema.get_field_index(time_field),
        time_field, floored_dates)
    start_day = pc.min(floored_dates).as_py()
    end_day = pc.max(floored_dates).as_py()
    if time_period == 'week':
        num_time_periods = (end_day - start_day).days // 7
    elif time_period == 'month':
        num_time_periods = (end_day.year - start_day.year) * 12
        num_time_periods += (end_day.month - start_day.month)
    else: # feature time_period == day
        num_time_periods = (end_day - start_day).days
    num_time_periods += 1 # end date is inclusive
    return wrangled_table, num_time_periods, start_day

def _compute_metadata(feature_table, cols, con):
    metadata = dict()
    metadata['fields'] = dict()
    metadata['event_col'] = cols[0]
    metadata['join_field'] = cols[1]
    metadata['time_field'] = 'time_period'
    m = metadata['fields']
    for col in feature_table.schema:
        m[col.name] = {'dtype' : str(col.type)}
        if _is_numeric(str(col.type)):
            min_max = pc.min_max(feature_table[col.name])
            m[col.name]['min'] = min_max['min'].as_py()
            m[col.name]['max'] = min_max['max'].as_py()
        else: 
            distincts = con.execute(
                f"SELECT DISTINCT {col.name} FROM feature_table").arrow()
            distincts_py = [val.as_py() for val in distincts[col.name]]
            m[col.name]['distincts'] = distincts_py
    return metadata

def _compute_labels(con, feature_table, join_field, label_col, label_values,
    n_time_periods_per_row_for_target):
    if len(label_values) == 1 and label_values[0] == None:
        col_names = [f"{label_col}_sum"]
    else:
        col_names = map(lambda c: f"{c}_count", _gen_col_names(label_col, label_values))
    expr = ' + '.join(map(lambda c: ' + '.join([f"(LEAD({c}, {i}) OVER win)" \
        for i in range(1, n_time_periods_per_row_for_target + 1)]), col_names))
    return con.execute(f"""
                            SELECT {join_field}, time_period, {expr} AS label 
                            FROM feature_table 
                            WINDOW win AS
                            (PARTITION BY {join_field} ORDER BY time_period ASC)
                        """).arrow()

def _compute_features(con, event_table, join_field, time_field,
    heavy_cat_cols, cat_cols, num_cols, preceding_time_periods):

    feat_names = []

    # Run value-enumeration for heavy cat cols
    cases = []
    for col, vals in heavy_cat_cols:
        col_names = _gen_col_names(col, vals)
        for e_sql, v_sql in zip(vals, col_names):
            feat_name_sql = f"{v_sql}_count"
            cases.append(f"""
                            CAST(SUM(CASE WHEN {col} = '{e_sql}'
                            THEN 1 ELSE 0 END) AS int64) AS {feat_name_sql}
                        """)
            feat_names.append(feat_name_sql)

    # Run lightweight categorical aggregations for cat columns
    cats = []
    for cat in cat_cols:
        cats.append(f"COUNT(DISTINCT {cat}) AS {cat}_count_distinct")
        cats.append(f"CAST(MODE({cat}) AS VARCHAR) AS {cat}_mode")
        feat_names.extend([f"{cat}_count_distinct", f"{cat}_mode"])

    # Run lightweight numeric aggregations for numeric columns
    conts = []
    for cont in num_cols:
        conts.append(f"CAST(SUM({cont}) AS double) AS {cont}_sum")
        conts.append(f"CAST(AVG({cont}) AS double) AS {cont}_avg")
        conts.append(f"CAST(STDDEV({cont}) AS double) AS {cont}_std")
        feat_names.extend([f"{cont}_sum", f"{cont}_avg", f"{cont}_std"])
    base_feats = ', '.join(cases + cats + conts)
    base_features = con.execute(f"""SELECT {join_field} AS {join_field}, {time_field} AS time_period,
                                    {base_feats} FROM event_table GROUP BY 1, 2;
                                """).arrow()
    final_feats = []
    for feat_name in feat_names:
        lags = ', '.join([f"LAG({feat_name}, {i}) OVER win AS {feat_name}_{i}" \
            for i in range(1, preceding_time_periods + 1)])
        final_feats.append(f"{feat_name}, {lags}")
    final_feats = ', '.join(final_feats)
    return con.execute(f"""SELECT {join_field}, time_period, {final_feats} FROM base_features WINDOW win AS
                        (PARTITION BY {join_field} ORDER BY time_period ASC);
                        """).arrow()


def _create_dummy_events(con, event_table, join_field, time_field,
    feature_time_period, start_day, num_time_periods):

    users = con.execute(f"""SELECT DISTINCT {join_field} AS {join_field}
                            FROM event_table
                         """).arrow()
    iters = pa.table({'iter': pa.array([i for i in range(num_time_periods)])})
    if feature_time_period == 'day':
        interval = "iter day"
    elif feature_time_period == 'week':
        interval = "(7 * iter) day"
    else: # 'month'
        interval = "iter month"
    dates = con.execute(f"""SELECT CAST(date_trunc('{feature_time_period}', DATE '{start_day}') +
                            INTERVAL {interval} AS timestamp) AS {time_field} FROM iters
                         """).arrow()
    return con.execute(f"""SELECT {join_field}, {time_field}
                           FROM users FULL JOIN dates ON 1 = 1
                        """).arrow()

def _process_top_k(top_k, field, _table, con):
    return con.execute(
        f"""
            SELECT {field}, COUNT({field}) AS cnt
            FROM _table
            GROUP BY {field}
            ORDER BY cnt DESC
            LIMIT {top_k};
        """).arrow()

def _process_cat_value_filter(top_k_vals, field, _table, con):
    top_k_vals_sql = ""
    for i,val in enumerate(top_k_vals):
        top_k_vals_sql += f"{field} = '{val}' "
        if i < len(top_k_vals) - 1:
            top_k_vals_sql += "OR "

    other = "__OTHER__"
    while other in top_k_vals:
        other += "_"
    other_sql = f"'{other}'"
    
    return con.execute(
        f"""
            SELECT CASE 
            WHEN {top_k_vals_sql} THEN {field} ELSE {other_sql} END
            AS {field} FROM _table;
        """).arrow()



from datetime import datetime
import json
import duckdb
import pyarrow.parquet as pq
import pyarrow.compute as pc
import pytorch_lightning as pl
from torch.utils.data import DataLoader
from torch.utils.data import random_split

# Local imports
from .ml.data import SupervisedDataset, _is_numeric
from .ml.models import *

def train(
    feature_table,
    metadata,
    model='linear',
    target='label',
    embed_join_field=True,
    embed_categoricals=False,
    embed_dim=32,
    batch_size=128,
    lr=1e-3,
    precision=32,
    max_epochs=100,
    num_loading_workers=0,
    val_mode=None,
    val_split=None,
    model_export_path=None,
    model_export_format='torch'):
    """
    Run an training job on a given feature_table. Accepts metadata as returned
    by train.

    Parameters
    __________
    feature_table : pl.LightningModule or str
        A table containing samples to train on (as returned by 
        featurize) or path to such a table. 

    metadata : dict
        A dictionary containing metadata as returned by train.

    model : str
        The model type to use

    target : label
        The name of the target column

    embed_join_field : bool
        Turn this flag on to embed the join field into embedding vectors that 
        are then passed into the model as input. 

    embed_categoricals : bool
        Turn this on to embed categorical columns into embedding vectors.

    embed_dim: int
        Embedding dimension 

    batch_size : int
        Batch size to use for inference job.

    lr : float
        Learning rate to use for gradient-based solvers.

    precision : int
        The floating-pint precision to use for inference job.

    num_loeading_workers : int
        The number of loading workers to use for job.

    Returns
    _______
    Predicted values (PyTorch.tensor). Returns values used to predict 
    """


    if type(feature_table) is str: # Assume this is a path
        feature_table = pq.read_table(feature_table)

    if type(metadata) is str: # Assume this is a path
        with open(metadata,'r') as M:
            metadata = json.load(M)

    # Set up embed_d variables to pass into Dataset
    embed_d_jk = embed_dim if embed_join_field else -1
    embed_d_cats = embed_dim if embed_categoricals else -1

    # Process validation split
    if val_mode == None:
        D = SupervisedDataset(feature_table, metadata, target, 
            embed_d_jk, embed_d_cats)
        D_train = D
        D_val = []
    elif val_mode == "random":
        assert type(val_split) == float
        D = SupervisedDataset(feature_table, metadata, target, 
            embed_d_jk, embed_d_cats)
        l = int(len(D)*val_split) 
        D_train, D_val = random_split(D, [len(D)-l, l])
    elif val_mode == "time":
        if type(val_split) == float:
            # Linear interpolate from first to last time period
            max_t = pc.max(feature_table[metadata['time_field']]).as_py()
            min_t = pc.min(feature_table[metadata['time_field']]).as_py()
            time_split_sql = str((1-val_split)*(max_t - min_t) + min_t)
        elif type(val_split) == datetime:
            time_split_sql = str(time_split_sql)
        else:
            raise ValueError
        
        # Time split is easier in SQL at table level
        con = duckdb.connect(database=':memory:')
        train_table = con.execute(
            f"""
                SELECT * FROM feature_table 
                WHERE {metadata['time_field']} <= '{time_split_sql}'
            """).arrow()
        val_table = con.execute(
            f"""
                SELECT * FROM feature_table 
                WHERE {metadata['time_field']} > '{time_split_sql}'
            """).arrow()

        D_train = SupervisedDataset(train_table, metadata, target, 
            embed_d_jk, embed_d_cats)
        D = D_train # Still want this var for consistency with rest of code

        D_val = SupervisedDataset(val_table, metadata, target,
            embed_d_jk, embed_d_cats)
    elif val_mode == 'join':
        raise NotImplementedError
    
    train_loader = DataLoader(D_train, batch_size=batch_size,
            num_workers=num_loading_workers)
    if len(D_val) > 0:
        val_loader = DataLoader(D_val, batch_size=batch_size,
            num_workers=num_loading_workers)
    else:
        val_loader = []

    # Instantiate model
    if model == 'linear':
        model = LinearRegression(*D.get_dims(),
            D.num_embedding_rows, embed_dim, lr=lr)
    elif model == 'logistic':
        model = LogisticRegression(*D.get_dims(),
            D.num_embedding_rows, embed_dim, lr=lr)
    elif model == 'naive_bayes':
        raise NotImplementedError
    else:
        raise ValueError(f'Model type {model} not supported.')

    # Train it, baby!
    trainer = pl.Trainer(precision=precision, max_epochs=max_epochs)
    trainer.fit(model, train_loader, val_loader)

    # Update metadata post-training
    new_metadata = dict()
    new_metadata['featurize'] = metadata
    new_metadata['train'] = {'embed_dim_join_field' : embed_d_jk,
        'embed_dim_categoricals' : embed_d_cats, 'target_column' : target,
        'fields' : D_train.fields}
    metadata = new_metadata

    # Export model if requested
    if model_export_path != None:
        if model_export_format == 'torch':
            torch.save(model, model_export_path)
        elif model_export_format == 'onnx':
            bsize = 1
            x_dense = torch.randn(bsize, D.dense_len, requires_grad=True)
            x_offs = torch.randint(1,(bsize, D.offset_len))
            torch.onnx.export(
                model,
                ((x_dense, x_offs),),
                model_export_path,
                input_names = ['input'],
                output_names = ['output'],
                dynamic_axes={'input' : {0 : 'bsize'},
                            'output' : {0 : 'bsize'}})
    
    return model, metadata



import pyarrow.compute as pc
import numpy as np
import torch
import torch.nn.functional as F
from torch.utils.data import Dataset

def _is_numeric(dtype):
    rtn = 'int' in dtype
    rtn = rtn or 'float' in dtype
    rtn = rtn or 'double' in dtype
    return rtn

def _bin(array, binning, n_bins):
    '''
    Converts a floating-point pyarrow array into categorical bins based on 
    quantiles. 
    '''
    if binning == 'linear':
        bins = np.arange(n_bins)
    elif binning == 'quantile':
        bins = np.quantile(array, q=[i*1/n_bins for i in range(n_bins+1)])
    return bins

class SupervisedDataset(Dataset):
    def __init__(self, 
        table, 
        metadata, 
        target_column, 
        embed_d_jk, # -1 if no embed => default to 1-hot
        embed_d_cats, # -1 if no embed => default to 1-hot
        inference=False,
        binning=None,
        n_bins=4
        ):

        self.table = table
        self.metadata = metadata
        self.m_fields = metadata['fields']
        self.dense_cols = []
        self.offset_cols = []
        self.target_column = target_column
        self.fields_dense = []
        self.fields_offs = []
        self.field2dim = dict()
        self.field2val2idx = dict()
        self.num_embedding_rows = 0
        self.target_dim = None
        self.dense_len = 0
        self.inference = inference

        for i,field in enumerate(self.m_fields):
            #if field == self.target_column:
            #    continue
                
            if _is_numeric(self.m_fields[field]['dtype']):
                array = np.nan_to_num(self.table[field].to_numpy(),copy=False)
                if binning == None:
                    self.dense_cols.append(
                        torch.tensor(array).float().unsqueeze(1))
                    self.field2dim[field] = 1
                    self.fields_dense.append(field)
                else:
                    bins = _bin(array, binning, n_bins)
                    binned_array = np.digitize(array, bins).astype(str)
                    distincts = np.unique(binned_array)
                    val2idx = {str(val):idx for idx,val in enumerate(distincts)}
                    N = len(val2idx) + 1 # Add one for OOV
                    self.field2val2idx[field] = val2idx
                    ten = torch.tensor(
                        [val2idx.get(val, N-1) for val in binned_array])
                    if embed_d_cats > 0: # Now we treat this as cat
                        self._process_embs(field, ten, N, embed_d_cats)
                        dim = embed_d_cats
                    else:
                        self._process_one_hots(field, ten, N, distincts)
                        dim = N
                    self.field2dim[field] = dim
                        
            else: # Should only be strings but need casting from pyarrow to str
                distincts = self.m_fields[field]['distincts']
                val2idx = {str(val):idx for idx,val in enumerate(distincts)}
                N = len(val2idx) + 1 # Add one for OOV
                self.field2val2idx[field] = val2idx
                if field == metadata['join_field']:
                    embed_d = embed_d_jk 
                else:
                    embed_d = embed_d_cats
                ten = torch.tensor(
                        [val2idx.get(
                            str(val), N-1) for val in self.table[field]])
                if embed_d > 0:
                    self._process_embs(field, ten, N, embed_d)
                    dim = embed_d
                else:
                    self._process_one_hots(field, ten, N, distincts)
                    dim = N
                self.field2dim[field] = dim

        if not inference:
            if _is_numeric(str(self.table.schema[self.table.column_names.index(
                    target_column)].type)):  
                self.target_col = torch.tensor(
                    self.table[target_column].to_numpy()).float().unsqueeze(1)
                self.target_dim = 1
            else:
                distincts = self.m_fields[field]['distincts']
                val2idx = {str(val):idx for idx,val in enumerate(distincts)}
                ten = torch.tensor(
                        [val2idx[str(val)] for val in self.table[field]])
                self.target_col = ten
                self.target_dim = len(val2idx)

        self.dense_cols = torch.concat(self.dense_cols, dim=1)
        self.offset_cols = torch.concat(self.offset_cols, dim=1)
        self.dense_len = self.dense_cols.shape[1]
        self.offset_len = self.offset_cols.shape[1]
        self.fields = self.fields_dense + self.fields_offs 

    def _process_embs(self, field, ten, N, embed_d):
        ten += self.num_embedding_rows
        ten = ten.unsqueeze(1)
        self.num_embedding_rows += N
        self.offset_cols.append(ten)
        self.fields_offs += [f"{field}_emb_{i}" for i in range(embed_d)]

    def _process_numerics(self, col, is_target=False):
        pass

    def _process_one_hots(self, field, ten, N, distincts, is_target=False):
        ten = F.one_hot(ten, num_classes=N)
        if not is_target:
            self.dense_cols.append(ten)
            self.fields_dense += [f"{field}_1hot_{val}" for val in distincts]
            self.fields_dense += [f"{field}_1hot__OOV_"]
        else:
            self.target_col = ten

    def get_dims(self):
        return sum([dim for dim in self.field2dim.values()]), self.target_dim

    def __len__(self):
        return len(self.table)

    def __getitem__(self, idx):
        x = (self.dense_cols[idx], self.offset_cols[idx])        
        rtn = x if self.inference else (x, self.target_col[idx])
        return rtn




import duckdb
import warnings
import pyarrow as pa
import pyarrow.parquet as pq
import pyarrow.compute as pc

def wrangle(
    join_field,
    time_field,
    primary_table_path=None,
    cols=None,
    event_col=None,
    table_paths=None,
    start_time=None,
    end_time=None):
    """
    Combine multiple event tables into a merged event table.

    Parameters
    __________
    join_field : str
        Name of column containing entitites to compute features over (e.g. user ID),
        must be present in all event tables.

    time_field : str
        Name of timestamp column, must be present in all event tables.

    primary_table_path : str, default None
        Path to Parquet file with the primary event table, which must also contain
        a column (`event_col`) with the type of each event. If None, attempt to use the
        first path in `table_paths`.

    cols : list of str, default None
        List of column names besides `join_field`, `time_field`, and `event_col` (for the
        primary table) to keep in the merged event table. Column names must be unique across
        the event tables.

    event_col : str, default None
        Name of the column with the event type in the primary event table. If None, will take
        the first item in `cols`.

    table_paths : list of tuple of str, str, default None
        List of tuples of (event type to give to all events in this table, path to Parquet
        file with event table).

    start_time : datetime, default None
        Filter for events with timestamp greater than this.

    end_time : datetime, default None
        Filter for events with timestamp less than this.

    Returns
    _______
    Merged event table (pyarrow.Table).
    """

    if cols == None:
        cols = []
    if table_paths == None:
        table_paths = []

    # We require an event col
    assert event_col != None or len(cols) >= 1, "We require an event col"

    # Process the event col
    print("Drifto.Wrangle: Processing event col...")
    if event_col == None:
        event_col = cols[0]
    else:
        cols.insert(0, event_col)

    # Process the table paths
    print("Drifto.Wrangle: Processing table paths...")
    if primary_table_path == None:
        primary_table_path = table_paths[0]
    else:
        table_paths.insert(0, ("primary_table", primary_table_path))

    join_schemas = list(map(lambda p: _read_schema(p[1]), table_paths))

    # Check core assertions
    print("Drifto.Wrangle: Checking core assertions...") 
    assert event_col in join_schemas[0].names, "Event col not found in primary table"
    for i, ((_, path), js) in enumerate(zip(table_paths, join_schemas)):
        assert join_field in js.names, f"Join field not found in {path}"
        assert time_field in js.names, f"Time field not found in {path}"
        if i > 0:
            assert not event_col in js.names, f"Event col found in {path}"

    cols_to_read = [{join_field, time_field} for i in range(len(join_schemas))]
    cols_to_read[0].add(event_col)
    json_queries = []
    json_cols_to_remove = set()

    # Try to find additional cols
    print("Drifto.Wrangle: Processing columns...") 
    for column in cols:
        col = column.split('->')
        found = False
        for i, js in enumerate(join_schemas):
            if col[0] in js.names:
                if found:
                    warnings.warn(f'Column {col[0]} found in multiple tables.')
                cols_to_read[i].add(col[0])
                found = True
        if not found:
            warnings.warn(f'Column {col[0]} not found in any table.')
        if len(col) > 1: # JSON field
            json_queries.append(column)
            # We don't want any JSON-valued cols in our analysis
            json_cols_to_remove.add(col[0])
    
    tables = []
    filters = []
    if start_time:
        filters.append((time_field, '>=', start_time))
    if end_time:
        filters.append((time_field, '<=', end_time))
    if len(filters) == 0:
        filters = None
    for cols, (_, path) in zip(cols_to_read, table_paths):
        tables.append(pq.read_table(
            path,
            columns=list(cols),
            filters=filters))

    # Always assume join_field is categorical and therefore can cast to str
    for i in range(len(tables)):
        idx = tables[i].column_names.index(join_field)
        tables[i] = tables[i].set_column(
            idx, join_field, pc.cast(tables[i][join_field], pa.string()))
        if i > 0:
            # Add primary column to secondary tables...
            # Filled with just the table name
            tables[i] = tables[i].append_column(event_col,
                pa.nulls(len(tables[i]),
                    pa.string()).fill_null(table_paths[i][0]))

    full_table = pa.concat_tables(tables, promote=True)

    print("Drifto.Wrangle: Connecting to DuckDB...")
    con = duckdb.connect(database=':memory:')
    print("Drifto.Wrangle: Finalizing feature table...")
    for jq in json_queries:
        clean = _clean_name(jq)
        res = con.execute(f"SELECT {jq} FROM full_table").arrow()
        full_table = full_table.append_column(clean, res[0])
    full_table = full_table.drop(list(json_cols_to_remove))

    print("Drifto.Wrangle: Done!")
    return full_table

def _clean_name(orig_name):
    name = ''.join([i if i.isalnum() else '_' for i in orig_name])
    # remove duplicate underscores
    name = name.lstrip('_').rstrip('_')
    return ''.join([c for i, c in enumerate(name) if c != '_' or name[i - 1] != '_'])

def _read_schema_remote(path):
    s3 = pa.fs.S3FileSystem()
    return pq.ParquetDataset(path, filesystem=s3).schema

def _read_schema(path):
    if path[:5] == 's3://':
        return _read_schema_remote(path[5:])
    else:
        return pq.read_schema(path)



from setuptools import setup

setup(
    name='drifto',
    version='0.1.2',
    license='ELv2',
    description='Automatic featurization and ML for event analytics',
    author='Drifto Technologies Inc',
    author_email='founders@driftoml.com',
    packages=['drifto', 'drifto.ml'],
    url = 'https://github.com/drifto-ml/drifto',
    keywords = ['data', 'machine learning', 'feature engineering', 'analytics', 'autoML'],
    install_requires=['pyarrow',
                      'duckdb',
                      'torch',
                      'pytorch_lightning',
                      'onnx']
)



import json
import pyarrow.parquet as pq
import pytorch_lightning as pl
from torch.utils.data import DataLoader

# Local imports
from .ml.data import SupervisedDataset

def inference(
    model,
    inference_table,
    metadata,
    batch_size=128,
    precision=32,
    num_loading_workers=0):
    """
    Run an inference job on a given inference_table. Accepts metadata as returned by train.

    Parameters
    __________
    model : pl.LightningModule or str
        A trained model (as returned by train) or a path to an exported model. 

    inference_table : pyarrow.Table or str
        A table containing samples to run inference on (as returned by 
        featurize) or path to such a table.

    metadata : dict
        A dictionary containing metadata as returned by train.

    batch_size : int
        Batch size to use for inference job.

    precision : int
        The floating-pint precision to use for inference job.

    num_loeading_workers : int
        The number of loading workers to use for job.

    Returns
    _______
    Predicted values (PyTorch.tensor). Returns values used to predict 
    """

    if type(inference_table) is str: # Assume this is a path
        inference_table = pq.read_table(inference_table)

    if type(metadata) is str: # Assume this is a path
        with open(metadata,'r') as M:
            metadata = json.load(M)

    embed_d_jk = metadata['train']['embed_dim_join_field']
    embed_d_cats = metadata['train']['embed_dim_categoricals']
    target = metadata['train']['target_column']
    D = SupervisedDataset(inference_table, metadata['featurize'], 
        target, embed_d_jk, embed_d_cats, inference=True)

    inference_loader = DataLoader(D, batch_size=batch_size,
            num_workers=num_loading_workers)

    trainer = pl.Trainer(precision=precision, max_epochs=1)
    predicts = trainer.predict(model, inference_loader)
    return predicts



import torch
from torch import nn
from torch.nn import functional as F
import pytorch_lightning as pl

class LinearRegression(pl.LightningModule):
    def __init__(self, 
        input_dim,
        output_dim,
        n_embeds,
        embed_dim,
        lr=1e-4):

        super().__init__()
        self.weights = nn.Linear(input_dim, output_dim)
        self.E = torch.nn.Embedding(n_embeds, embed_dim)
        self.lr = lr

    def forward(self, x):
        x_dense, x_offs = x
        x = torch.concat((x_dense, self.E(x_offs).squeeze(1)),dim=1)
        return self.weights(x)

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.lr)
        return optimizer

    def training_step(self, train_batch, batch_idx):
        x, y = train_batch
        loss = F.mse_loss(self(x), y)
        self.log('train_loss', loss)
        return loss

    def validation_step(self, val_batch, batch_idx):
        pass

    def predict_step(self, batch, batch_idx):
        x = batch
        return self(x)

class LogisticRegression(LinearRegression):
    def __init__(self, 
        input_dim,
        output_dim,
        n_embeds,
        embed_dim,
        lr=1e-4):
        super().__init__(input_dim, output_dim, n_embeds, embed_dim, lr=lr)

    def forward(self, x):
        return F.softmax(super().forward(x))

    def training_step(self, train_batch, batch_idx):
        x, y = train_batch
        loss = F.cross_entropy(self(x), y)
        self.log('train_loss', loss)
        return loss

class NaiveBayes(pl.LightningModule):
    def __init__(self,
        input_dim : list[int],
        n_cats : list[int],
        n_classes: int):
        super().__init__()
        self.p = torch.zeros()
        self.p_lbl = torch.zeros()
        self.p_num = None
        self.n_classes = n_classes
        L = sum(n_cats)
        self.offsets = torch.tensor(n_cats) - 1
        self.N = torch.nn.Parameter(torch.zeros(L))
        self.C = torch.nn.Parameter(torch.zeros(n_classes))
        self.P = torch.nn.Parameter(torch.zeros(L, n_classes))

    def forward(self, x, include_prob=False):
        # TODO: Consider converting this into a log sum for numerical reasons
        Pi = torch.prod(self.P[x]/self.N[x], dim=1) # TODO dim
        y_hat = torch.argmax(self.C * Pi)
        if include_prob:
            return y_hat
        else:
            Z = torch.sum(self.C * Pi)
            return y_hat, self.C * Pi / Z

    def configure_optimizers(self):
        optimizer = torch.optim.Optimizer
        optimizer.step = lambda : None
        return optimizer

    def training_step(self, train_batch, batch_idx):
        x, y = train_batch
        loss = F.cross_entropy(self(x), y)
        
        # Perform update
        self.N[x] += 1
        self.P[torch.stack(x,y)] += 1 # TODO
        self.C += torch.sum(F.one_hot(y, num_classes=self.n_classes), dim=0)
        self.log('train_loss', loss)
        return loss

    def validation_step(self, val_batch, batch_idx):
        pass


